@@ -1,3 +1,21 @@
+# Copyright (c) 2022 Kyle R. Lennon (kyle.lennon08@gmail.com).
+#
+# This file is part of mastercurves.
+#
+# mastercurves is free software: you can redistribute it and/or modify it under the terms of the
+# GNU General Public License as published by the Free Software Foundation, either version 3
+# of the License, or (at your option) any later version.
+#
+# mastercurves is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
+# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
+# See the GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License along with mastercurves.
+# If not, see <https://www.gnu.org/licenses/>.
+#
+# Please cite "A Data-Driven Method for Automated Data Superposition with Applications
+# in Soft Matter Science" (https://arxiv.org/abs/2204.09521) if you use any part of the code.
+
 import numpy as np
 import matplotlib.pyplot as plt
 from matplotlib.colors import to_hex
@@ -9,12 +27,84 @@ import numdifftools as nd
 import random
 
 class MasterCurve:
-    """
+    r"""
     Class definition for a master curve, consisting of multiple data sets superimposed.
+
+    A :attr:`MasterCurve` object will contain all of the data used to construct a master
+    curve, along with the coordinate transformations, parameters for coordinate transformations
+    (such as shift factors), associated uncertainties, Gaussian process models, and
+    transformed data.
+
+    Attributes:
+        :attr:`xdata` (:attr:`list[array_like]`): list whose elements are the independent
+        coordinates for a data set at a given state
+
+        :attr:`ydata` (:attr:`list[array_like]`): list whose elements are the dependent
+        coordinates for a data set at a given state
+
+        :attr:`states` (:attr:`list[float]`): list whose elements are the value of the state
+        defining each data set
+
+        :attr:`htransforms` (:attr:`list[Transform]`): list whose elements are the
+        coordinate transforms performed on the independent coordinate (i.e. horizontally)
+
+        :attr:`hparam_names` (:attr:`list[string]`): list whose elements are the names
+        of the parameters for the horizontal transforms
+
+        :attr:`hbounds` (:attr:`list[tuple]`): list whose elements are the upper and lower
+        bounds for the horizontal transformation parameters
+
+        :attr:`hshared` (:attr:`list[bool]`): list whose elements indicate whether the
+        corresponding element of :attr:`hparam_names` is a parameter whose value is shared
+        across all states (:attr:`True`) or takes on an independent value for each state
+        (:attr:`False`)
+
+        :attr:`hparams` (:attr:`list[list[float]]`): list whose elements are the values of the
+        horizontal transformation parameters at each state
+
+        :attr:`huncertainties` (:attr:`list[float]`): list whose elements are the uncertainties
+        associated with the values in :attr:`hparams`
+
+        :attr:`vtransforms` (:attr:`list[Transform]`): list whose elements are the
+        coordinate transforms performed on the dependent coordinate (i.e. vertically)
+
+        :attr:`vparam_names` (:attr:`list[string]`): list whose elements are the names
+        of the parameters for the vertical transforms
+
+        :attr:`vbounds` (:attr:`list[tuple]`): list whose elements are the upper and lower
+        bounds for the vertical transformation parameters
+
+        :attr:`vshared` (:attr:`list[bool]`): list whose elements indicate whether the
+        corresponding element of :attr:`vparam_names` is a parameter whose value is shared
+        across all states (:attr:`True`) or takes on an independent value for each state
+        (:attr:`False`)
+
+        :attr:`vparams` (:attr:`list[list[float]]`): list whose elements are the values of the
+        vertical transformation parameters at each state
+
+        :attr:`vuncertainties` (:attr:`list[float]`): list whose elements are the uncertainties
+        associated with the values in :attr:`vparams`
+
+        :attr:`kernel` (:attr:`sklearn.gaussian_process.kernels.Kernel`): kernel function
+        for the Gaussian process model
+
+        :attr:`gps` (:attr:`list[sklearn.gaussian_process.GaussianProcessRegressor]`): list
+        whose elements are Gaussian process models for each state
+
+        :attr:`xtransformed` (:attr:`list[array_like]`): list whose elements are the transformed
+        independent coordinates for a data set at a given state
+
+        :attr:`ytransformed` (:attr:`list[array_like]`): list whose elements are the transformed
+        dependent coordinates for a data set at a given state
     """
     def __init__(self, fixed_noise=0.04):
-        """
-        Initialize object with some x_data sets and corresponding y_data sets.
+        r"""
+        Initialize a MasterCurve object.
+
+        Args:
+            :attr:`fixed_noise` (:attr:`float`): the fixed noise level for the Gaussian process
+            models, corresponding to experimental uncertainty not evident in the data. By default,
+            the noise level is 0.04.
         """
         self.xdata = []
         self.xtransformed = []
@@ -42,21 +132,25 @@ class MasterCurve:
         self.vshared = []
 
     def add_data(self, xdata_new, ydata_new, states_new):
-        """
-        Add a data set to the master curve.
-        Inputs:
-            xdata (list) - x-data sets corresponding to one or more states.
-                            Each element in the list corresponds to the x-data at a given state.
-            ydata (list) - y-data sets corresponding to one or more states.
-                            Each element in the list corresponds to the y-data at a given state.
-            states (list) - values of the state parameters corresponding to the elements in x_data and y_data.
+        r"""
+        Add a data set or data sets to the master curve.
+
+        Args:
+            :attr:`xdata_new` (:attr:`array_like` or :attr:`list[array_like]`): array(s) corresponding to
+            the dependent coordinates for given states
+
+            :attr:`ydata_new` (:attr:`array_like` or :attr:`list[array_like]`): array(s) corresponding to
+            the independent coordinates for given states
+
+            :attr:`states_new` (:attr:`float` or :attr:`list[float]`): values of the state parameter
+            corresponding to the data in :attr:`xdata_new` and :attr:`ydata_new`
         """
         self.xdata += xdata_new
         self.ydata += ydata_new
         self.states += states_new
 
     def clear(self):
-        """
+        r"""
         Clear the master curve's data (useful for memory management).
         """
         self.xdata = []
@@ -77,18 +171,22 @@ class MasterCurve:
         self.vshared = []
 
     def set_gp_kernel(self, kernel):
-        """
+        r"""
         Set the kernel function for the Gaussian Processes used to fit the data.
-        Inputs:
-            kernel (from sklearn.gaussian_process.kernels) - the (potentially composite) kernel function.
+
+        Args:
+            :attr:`kernel` (:attr:`sklearn.gaussian_process.kernels.Kernel`): the
+            (potentially composite) kernel function
         """
         self.kernel = kernel
 
     def _fit_gps(self, overwrite=False):
-        """
+        r"""
         Fit Gaussian Process models to each data set in this master curve.
-        Inputs:
-            overwrite (Boolean) - overwrite existing GP fits if any exist (when True).
+
+        Args:
+            :attr:`overwrite` (:attr:`bool`): overwrite existing GP fits if any exist (when :attr:`True`).
+            By defailt, it is :attr:`False`.
         """
         # Clear GPs if we are overwriting
         if overwrite:
@@ -105,10 +203,12 @@ class MasterCurve:
             self.gps += [GaussianProcessRegressor(kernel=self.kernel).fit(x.reshape(-1,1), y)]
 
     def add_htransform(self, htransform):
-        """
-        Add a horizontal transformation (or series of transformations) to the master curve (sequantially).
-        Inputs:
-            htransform - transformation object
+        r"""
+        Add a horizontal transformation (or series of sequential transformations) to the master curve.
+
+        Args:
+            :attr:`htransform` (:attr:`Transform`): object of a Transform class, which implements the
+            coordinate transformation and stores information about transformation parameters
         """
         self.htransforms += [htransform]
         if htransform.type == "Multiply":
@@ -121,10 +221,12 @@ class MasterCurve:
         self.hshared += [htransform.shared]
 
     def add_vtransform(self, vtransform):
-        """
-        Add a vertical transformation (or series of transformations) to the master curve (sequantially).
-        Inputs:
-            vtransform - transformation object
+        r"""
+        Add a vertical transformation (or series of sequential transformations) to the master curve.
+
+        Args:
+            :attr:`vtransform` (:attr:`Transform`): object of a Transform class, which implements the
+            coordinate transformation and stores information about transformation parameters
         """
         self.vtransforms += [vtransform]
         if vtransform.type == "Multiply":
@@ -137,15 +239,27 @@ class MasterCurve:
         self.vshared += [vtransform.shared]
 
     def _change_states(self, transforms, data, s1, s2, p1, p2):
-        """
-        Given a set of transforms, transform the data coordinate from state s1 to state s2.
-        Inputs:
-            transforms - the list of transforms to take from a state to the reference
-            data - the data coordinate vector
-            s1, s2 - beginning state (s1) and final state (s2)
-            p1, p2 - transforming parameters corresponding to s1 and s2
-        Outputs:
-            transformed - transformed data from s1 to s2
+        r"""
+        Given a set of transforms, transform the data coordinate from state :attr:`s1` to state :attr:`s2`.
+
+        Args:
+            :attr:`transforms` (:attr:`list[Transform]`): the list of transforms to take data
+            from state :attr:`s1` to state :attr:`s2`
+
+            :attr:`data` (:attr:`array_like`): vector containing the coordinates to be transformed
+
+            :attr`s1` (:attr:`float`): the initial state
+
+            :attr`s2` (:attr:`float`): the final state
+
+            :attr:`p1` (:attr:list[`float`]): the values of the transform parameters corresponding
+            to each transform in :attr:`transforms`, for state :attr:`s1`
+
+            :attr:`p2` (:attr:list[`float`]): the values of the transform parameters corresponding
+            to each transform in :attr:`transforms`, for state :attr:`s2`
+
+        Returns:
+            :attr:`transformed` (:attr:`array_like`): vector containing the transformed coordinate
         """
         transformed = data
         for i in range(len(transforms)):
@@ -155,15 +269,20 @@ class MasterCurve:
         return transformed
 
     def _unpack(self, ind):
-        """
-        Unpack the data associated with index ind.
-        Inputs:
-            ind - index in lists
-        Outputs:
-            x - x-coordinate data
-            y - y-coordinate data
-            s - state
-            gp - fit GP model
+        r"""
+        Unpack the data associated with index :attr:`ind`.
+
+        Args:
+            :attr:`ind` (:attr:`int`): index of the state
+
+        Returns:
+            :attr:`x` (:attr:`array_like`): corresponding x-coordinates
+
+            :attr:`y` (:attr:`array_like`): corresponding y-coordinates
+
+            :attr:`s` (:attr:`float`): corresponding state
+
+            :attr:`gp` (:attr:`sklearn.gaussian_process.GaussianProcessRegressor): corresponding GP model
         """
         x = self.xdata[ind]
         y = self.ydata[ind]
@@ -172,15 +291,25 @@ class MasterCurve:
         return x, y, s, gp
 
     def _pairwise_nlog_posterior(self, ind1, ind2, hp1, hp2, vp1, vp2, lamh, lamv):
-        """
-        Compute the pairwise negative log likelihood loss between two curves.
-        Inputs:
-            ind1, ind2 - the indices in the data lists corresponding to the two data sets.
-            hp1, hp2 - the horizontal shift parameters for each data set
-            vp1, vp2 - the vertical shift parameters for each data set
-            lamh, lamv - hyperparameter(s) defining the prior distributions over horizontal and vertical shifts
-        Outputs:
-            loss - the negative log of the posterior, computed over these two data sets.
+        r"""
+        Compute the pairwise negative log posterior loss between two curves.
+
+        Args:
+            :attr:`ind1`, :attr:`ind2` (:attr:`int`): the indices in the data lists
+            corresponding to the two data sets
+
+            :attr:`hp1`, :attr`hp2` (:attr:`list[float]`): the horizontal shift parameters
+            for each data set, corresponding to each coordinate transformation
+
+            :attr:`vp1`, :attr`vp2` (:attr:`list[float]`): the vertical shift parameters
+            for each data set, corresponding to each coordinate transformation
+
+            :attr:`lamh`, :attr:`lamv` (:attr:`list[float]`): hyperparameter(s) defining
+            the prior distributions over the corresponding horizontal and vertical shifts
+
+        Returns:
+            :attr:`loss` (:attr:`float`): the negative log of the posterior,
+            computed over these two data sets
         """
         # Get the data and GPs
         x1, y1, s1, gp1 = self._unpack(ind1)
@@ -215,15 +344,23 @@ class MasterCurve:
         return loss
 
     def _objective(self, inds, hps, vps, lamh, lamv):
-        """
-        Compute the loss over all data sets in inds, given params hps and vps.
-        Inputs:
-            inds - indices of data curves to shift
-            hps - horizontal shifting parameters
-            vps - vertical shifting parameters
-            lamh, lamv - the horizontal and vertical regularization hyperparameters
-        Outputs:
-            loss - the total loss
+        r"""
+        Compute the loss over all data sets in :attr:`inds`, given params :attr:`hps` and :attr:`vps`.
+
+        Args:
+            :attr:`inds` (:attr:`list[int]`): indices of data sets to shift
+
+            :attr:`hps` - (:attr:`list[float]` or :attr:`list[list[float]]`): horizontal
+            shifting parameters for each state in :attr:`inds` and each coordinate transformation
+
+            :attr:`vps` - (:attr:`list[float]` or :attr:`list[list[float]]`): vertical
+            shifting parameters for each state in :attr:`inds` and each coordinate transformation
+
+            :attr:`lamh`, :attr:`lamv` (:attr:`list[float]`) the horizontal and vertical
+            regularization hyperparameters
+
+        Returns:
+            :attr:`loss` (:attr:`float`): the total loss
         """
         # For every sequential pair in inds
         loss = 0
@@ -289,12 +426,20 @@ class MasterCurve:
         return loss
 
     def superpose(self, lamh=None, lamv=None):
-        """
+        r"""
         Optimize the transformations to superpose the data sets onto a single master curve.
-        Inputs:
-            lamh, lamv - regularization hyperparameters for vertical and horizontal shifts
-        Outputs:
-            loss_min - the minimum loss computed during superposition (either total, or pairwise)
+
+        Args:
+            :attr:`lamh` (:attr:`list[float]`): hyperparameters for the horizontal shifts,
+            corresponding to each state in the master curve. Defaults to :attr:`None`, meaning
+            a uniform (unregularized) prior.
+
+            :attr:`lamv` (:attr:`list[float]`): hyperparameters for the vertical shifts,
+            corresponding to each state in the master curve. Defaults to :attr:`None`, meaning
+            a uniform (unregularized) prior.
+
+        Returns:
+            :attr:`loss_min` (:attr:`float`): the minimum loss computed during superposition
         """
         # First, fit the GPs
         self._fit_gps()
@@ -419,6 +564,7 @@ class MasterCurve:
                                 vp1 = self.vparams[0][ind1]
                                 vp2 = vp1*np.exp(mu_prod)
                                 self.vparams[0][ind2] = vp2
+                                self.vuncertainties[0][ind2] = vp1*np.exp(mu_prod)*np.sqrt(sig_prod)
         else:
             # Perform pairwise shifting
             loss_min = []
@@ -519,16 +665,18 @@ class MasterCurve:
                                 vp1 = self.vparams[0][ind1]
                                 vp2 = vp1*np.exp(mu_prod)
                                 self.vparams[0][ind2] = vp2
+                                self.vuncertainties[0][ind2] = vp1*np.exp(mu_prod)*np.sqrt(sig_prod)
 
         # Return error if more than one parameter is specified (other than vertical log Multiply)
         if vparams[-1] > 1:
             raise ValueError("Too many transformations provided. Current version supports one tranformation (in addition to one vertical log Multiply())")
 
         self._shift_data()
+        self._propagate_uncertainty()
         return loss_min
 
     def _shift_data(self):
-        """
+        r"""
         Save the shifted data for this master curve.
         """
         self.xtransformed = []
@@ -551,12 +699,53 @@ class MasterCurve:
                 y = self.vtransforms[l].forward(p, self.states[k], y)
             self.ytransformed += [y]
 
-    def plot(self, log=True, colormap=plt.cm.tab10, colorby="index"):
+    def _propagate_uncertainty(self):
+        r"""
+        For Multiply transforms, propagate uncertainty estimates in the shift factors
+        so that the uncertainties are cumulative.
         """
+        # Only supported when there is one coordinate transform per axis
+        if len(self.htransforms) == 1:
+            if self.htransforms[0].type == "Multiply":
+                huncertainties_new = [0]
+                for i in range(1, len(self.huncertainties[0])):
+                    huncertainties_new += [self.hparams[0][i] *np.sqrt((self.huncertainties[0][i]/self.hparams[0][i])**2 +
+                                (huncertainties_new[i-1]/self.hparams[0][i-1])**2)]
+                self.huncertainties[0] = huncertainties_new
+
+        # Now vertical uncertainties
+        if len(self.vtransforms) == 1:
+            if self.vtransforms[0].type == "Multiply":
+                vuncertainties_new = [0]
+                for i in range(1, len(self.vuncertainties[0])):
+                    vuncertainties_new += [self.vparams[0][i] *np.sqrt((self.vuncertainties[0][i]/self.vparams[0][i])**2 +
+                                (vuncertainties_new[i-1]/self.vparams[0][i-1])**2)]
+                self.vuncertainties[0] = vuncertainties_new
+
+    def plot(self, log=True, colormap=plt.cm.tab10, colorby="index"):
+        r"""
         Plot the data, GPs, and master curve.
-        Inputs:
-            log - if True, the data is the log of the desired quantity, so plot in linear space
-            colormap - colormap for plotting curves with different states.
+
+        Args:
+            :attr:`log` (:attr:`bool`): whether the data represents the logarithm of the
+            measured quantity. Defaults to :attr:`True`.
+
+            :attr:`colormap` (:attr:`matplotlib.colors.Colormap`): colormap for plotting.
+            Defaults to the tab10 colormap.
+
+            :attr:`colorby` (:attr:`string`): how to color the data. Options are :attr:`index`
+            for coloring by index, or :attr:`state` for coloring by the value of the state.
+            Defaults to :attr:`index`.
+
+        Returns:
+            :attr:`fig1`, :attr:`ax1` (:attr:`matplotlib.Figure` and :attr:`matplotlib.axes.Axes`): the
+            figure and axes objects displaying the raw (untransformed data)
+
+            :attr:`fig2`, :attr:`ax2` (:attr:`matplotlib.Figure` and :attr:`matplotlib.axes.Axes`): the
+            figure and axes objects displaying the untransformed data and GP models
+
+            :attr:`fig3`, :attr:`ax3` (:attr:`matplotlib.Figure` and :attr:`matplotlib.axes.Axes`): the
+            figure and axes objects displaying the superposed data (i.e. the master curve)
         """
         # Plot the data
         fig1, ax1 = plt.subplots(1,1)
@@ -612,34 +801,96 @@ class MasterCurve:
         return fig1, ax1, fig2, ax2, fig3, ax3
 
     def change_ref(self, ref_state, a_ref=1, b_ref=1):
-        """
-        Change the reference state.
-        Inputs:
-            ref_state - the new reference state, which may be one of the current states or not
-            a_ref - if ref_state is not a current state, must be provided. This is the new reference's horizontal shift
-                    with respect to the current reference
-            b_ref - if ref_state is not a current state, must be provided. This is the new reference's vertical shift
-                    with respect to the current reference
+        r"""
+        Change the reference state for the master curve.
+
+        Args:
+            :attr:`ref_state` (:attr:`float`): the new reference state, which may or may not
+            be one of the current states
+
+            :attr:`a_ref` (:attr:`float`): if :attr:`ref_state` is not a current state,
+            must be provided. This is the new reference's horizontal shift with respect
+            to the current reference. Defaults to 1.
+
+            :attr:`b_ref` (:attr:`float`): if :attr:`ref_state` is not a current state,
+            must be provided. This is the new reference's vertical shift with respect
+            to the current reference. Defaults to 1.
         """
         if len(self.htransforms) == 1:
             if self.htransforms[0].type == "Multiply":
                 if ref_state in self.states and a_ref == 1:
-                    a_ref = self.hparams[0][self.states.index(ref_state)]
+                    # Uncertainty propagation
+                    ind_ref = self.states.index(ref_state)
+                    if self.huncertainties[0][0] != 0:
+                        print("Warning: reference state has already been changed. Uncertainty estimates may no longer be accurate")
+                    else:
+                        huncertainties_pairwise = [0]
+                        for i in range(1, len(self.huncertainties[0])):
+                            huncertainties_pairwise += [self.hparams[0][i]
+                                    *np.sqrt((self.huncertainties[0][i]/self.hparams[0][i])**2
+                                        - (self.huncertainties[0][i-1]/self.hparams[0][i-1])**2)]
+                        self.huncertainties[0][ind_ref] = 0
+                        for i in range(ind_ref-1, -1, -1):
+                            self.huncertainties[0][i] = (self.hparams[0][i]
+                                    *np.sqrt((huncertainties_pairwise[i]/self.hparams[0][i])**2
+                                        + (self.huncertainties[0][i+1]/self.hparams[0][i+1])**2))
+                        for i in range(ind_ref+1, len(self.huncertainties[0])):
+                            self.huncertainties[0][i] = (self.hparams[0][i]
+                                    *np.sqrt((huncertainties_pairwise[i]/self.hparams[0][i])**2
+                                        + (self.huncertainties[0][i-1]/self.hparams[0][i-1])**2))
+
+                    # Find the reference shift factor
+                    a_ref = self.hparams[0][ind_ref]
+
+                # Rescale the shift factors and uncertainties
                 for k in range(len(self.hparams[0])):
                     self.hparams[0][k] = self.hparams[0][k]/a_ref
+                    self.huncertainties[0][k] = self.huncertainties[0][k]/a_ref
+
+        # Now vertical shifts
         if len(self.vtransforms) == 1:
             if self.vtransforms[0].type == "Multiply":
                 if ref_state in self.states and b_ref == 1:
+                    # Uncertainty propagation
+                    ind_ref = self.states.index(ref_state)
+                    if self.vuncertainties[0][0] != 0:
+                        print("Warning: reference state has already been changed. Uncertainty estimates may no longer be accurate")
+                    else:
+                        vuncertainties_pairwise = [0]
+                        for i in range(1, len(self.vuncertainties[0])):
+                            vuncertainties_pairwise += [self.vparams[0][i]
+                                    *np.sqrt((self.vuncertainties[0][i]/self.vparams[0][i])**2
+                                        - (self.vuncertainties[0][i-1]/self.vparams[0][i-1])**2)]
+                        self.vuncertainties[0][ind_ref] = 0
+                        for i in range(ind_ref-1, -1, -1):
+                            self.vuncertainties[0][i] = [self.vparams[0][i]
+                                    *np.sqrt((vuncertainties_pairwise[i]/self.vparams[0][i])**2
+                                        + (self.vuncertainties[0][i+1]/self.vparams[0][i+1])**2)]
+                        for i in range(ind_ref+1, len(self.vuncertainties[0])):
+                            self.vuncertainties[0][i] = [self.vparams[0][i]
+                                    *np.sqrt((vuncertainties_pairwise[i]/self.vparams[0][i])**2
+                                        + (self.vuncertainties[0][i-1]/self.vparams[0][i-1])**2)]
+
+                    # Find the reference shift factor
                     b_ref = self.vparams[0][self.states.index(ref_state)]
+
+                # Rescale the shift factors and uncertainties
                 for k in range(len(self.vparams[0])):
                     self.vparams[0][k] = self.vparams[0][k]/b_ref
+                    self.vuncertainties[0][k] = self.vuncertainties[0][k]/b_ref
         self._shift_data()
 
     def output_table(self, file=None):
-        """
-        Write a csv file with a table of all parameters (or return a data frame if no file given).
-        Inputs:
-            file - the (optional) file to which the data frame is written
+        r"""
+        Write a csv file with a table of all parameters (and return as a data frame).
+
+        Args:
+            :attr:`file` (:attr:`string`): (optional) path to the file to which
+            the data frame will be written
+
+        Returns:
+            :attr:`df` (:attr:`pandas.DataFrame`): data frame containing the transformation
+            parameters
         """
         data = {"state": self.states}
         for i in range(len(self.htransforms)):
@@ -660,16 +911,26 @@ class MasterCurve:
         return df
 
     def _mccv(self, x1, y1, s1, x2, y2, s2, lamh, lamv, alpha, folds):
-        """
+        r"""
         Compute the score of a Monte-Carlo cross validation test.
-        Inputs:
-            x1, y1, s1 - the first data set
-            x2, y2, s2 - the second data set
-            lamh, lamv - horizontal and vertical hyperparameters for the prior distribution
-            alpha - the fraction of points kept in the training set
-            folds - the number of MCCV folds to average over
-        Outputs:
-            score - the MCCV score (averaged over folds)
+
+        Args:
+            :attr:`x1`, :attr:`y1`, :attr:`s1` (:attr:`array_like`, :attr:`array_like`,
+            :attr:`float`): the independent coordinates, dependent coordinates, and state
+            for the first data set
+
+            :attr:`x2`, :attr:`y2`, :attr:`s2` (:attr:`array_like`, :attr:`array_like`,
+            :attr:`float`): the independent coordinates, dependent coordinates, and state
+            for the second data set
+
+            :attr:`lamh`, :attr:`lamv` (:attr:`list[float]`, :attr:`list[float]`): horizontal
+            and vertical hyperparameters for the prior distributions over parameters
+
+            :attr:`alpha` (:attr:`float`): the fraction of points kept in the training set
+            :attr:`folds` (:attr:`int`): the number of MCCV folds
+
+        Returns:
+            :attr:`score` (:attr:`float`): the MCCV score (averaged over folds)
         """
         scores = []
         for f in range(folds):
@@ -747,15 +1008,28 @@ class MasterCurve:
         return average, sterr
 
     def hpopt(self, lamh=None, lamv=None, npoints=100, alpha=0.5, folds=10):
-        """
+        r"""
         Perform hyperparameter optimization on the prior (regularization) using MCCV.
-        Inputs:
-            lamh, lamv - ranges (tuples) for the hyperparameter search. If not searching this parameter, then None
-            npoints - number of grid points to search
-            alpha - keep rate for MCCV
-            folds - number of MCCV folds
-        Outputs:
-            lamh_opt, lamv_opt - optimal hyperparameters
+
+        Args:
+            :attr:`lamh` (:attr:`list[tuple[float]]`): ranges (tuples) for the horizontal
+            hyperparameter search. If not searching this parameter, then the entry should be None.
+
+            :attr:`lamv` (:attr:`list[tuple[float]]`): ranges (tuples) for the vertical
+            hyperparameter search. If not searching this parameter, then the entry should be None.
+
+            :attr:`npoints` (:attr:`int`): number of grid points to search. Default is 100.
+
+            :attr:`alpha` (:attr:`float`): keep rate for MCCV. Default is 0.5.
+
+            :attr:`folds` (:attr:`int`): number of MCCV folds. Default is 10.
+
+        Returns:
+            :attr:`lamh_opt` (:attr:`list[float]`): optimal horizontal hyperparameters
+            at each state
+
+            :attr:`lamv_opt` (:attr:`list[float]`): optimal vertical hyperparameters
+            at each state
         """
         # Loop through each pair of data sets (only pairwise optimization supported)
         lamh_mins = []
